/*  pngrewrite    v1.0.0

  A quick and dirty utility to reduce unnecessarily large palettes 
  and bit depths in PNG image files.

  By Jason Summers <jason1@pobox.com> <http://www.pobox.com/~jason1/png/>
  This program and source code may be used without restriction.


  Pngrewrite will:

  * Remove any unused palette entries, and write a palette that is only as
    large as needed.

  * Remove (collapse) any duplicate palette entries.

  * Convert non-palette image (including grayscale) to palette images,
    provided they contain no more than 256 different colors.

  * Move any colors with transparency to the beginning of the palette, and
    write a tRNS chunk that is a small as possible.

  * Reduce the bit-depth (bits per pixel) as much as possible.

  All of this is just basic stuff that any respectable image writing program
  should always do automatically, but for some reason, many of them do not.


  Pngrewrite is intended mainly for use on images that will be used on web
  pages. It does not change the image; it only tries to reduce the file size.

  --WARNING--
  This version of pngrewrite removes nearly all extra (ancillary) information
  from the file, such as text comments. Although this does make the file size
  smaller, the removed information may sometimes be important. For that
  reason, you should only use pngrewrite on image that were created by you.

  The only ancillary chunks that are NOT removed are:
     gAMA  - Image gamma setting
     sRGB  - srgb color space indicator
     bKGD and tRNS - Background color and transparency are maintained. The
        actual chunk may be modified according to the new color structure.

  If the original image was interlaced, the new one will also be interlaced.

  Note that converting grayscale images to palette images is usually (but not
  quite always) a bad thing to do, and will probably increase the file size.

  Pngrewrite will not work at all on images that have more than 256 colors.
  Colors with the same RGB values but a different level of transparency
  count as different colors. The background color counts as an extra color
  if it does not occur in the image.
  
  It will also not work at all on images that have a color depth of 16 bits,
  since they cannot have a palette.

  This is a very inefficient program. It is (relatively) slow, and uses tons
  of memory. To be specific, it uses about 5 bytes per pixel, no matter what
  the bit depth of the image is.

  This program is (hopefully) reasonably portable, and should compile with
  most ANSI C compilers. It requires the libpng and zlib libraries.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#include <search.h> /* for qsort */
#endif

#include "png.h"

typedef struct {
	unsigned char red;
	unsigned char green;
	unsigned char blue;
	unsigned char alpha;
} tmppal_t;


tmppal_t tmppal[256];
int pal_used;

unsigned char *image1, *image2;
unsigned char **row_pointers;
int rowbytes, channels;
png_uint_32 width, height;
int bit_depth, color_type, interlace_type;
int ori_pal_size;

int has_gama=0;
int has_bkgd=0;
int has_srgb=0;

double image_gamma;
tmppal_t bkgd;             /* RGB background color */
unsigned char bkgd_pal;    /* new background color palette entry */
int srgb_intent;

int read_png(char *file_name)
{
	png_structp png_ptr;
	png_infop info_ptr;
	png_colorp ori_pal;
	int ori_bpp;

	FILE *fp;
	int i;
	png_color_16 *image_background;


	if ((fp = fopen(file_name, "rb")) == NULL) {
		printf("Can't open file %s\n",file_name);
		return 0;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL);

	if(png_ptr == NULL) {
		printf("Error creating read_struct\n");
		fclose(fp);
		return 0;
	}


	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return 0;
	}

	if (setjmp(png_ptr->jmpbuf)) {
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(fp);
		printf("Error reading file\n");
		return 0;
	}

	png_init_io(png_ptr, fp);

	png_read_info(png_ptr, info_ptr);

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

	if(bit_depth>8) {
		printf("This image can't be converted because it has 16 bits/sample.\n");
		goto abort1;
	}

	/* figure out bits per pixel so we can print it */
	ori_bpp= bit_depth*png_get_channels(png_ptr,info_ptr);

    if (color_type == PNG_COLOR_TYPE_PALETTE) {
		ori_pal_size=0;

		/* we only do this to get the palette size so we can print it */
		png_get_PLTE(png_ptr,info_ptr,&ori_pal,&ori_pal_size);

		printf("original palette size:   %3d, %2d bpp\n",ori_pal_size,ori_bpp);

        png_set_expand(png_ptr);
	}
	else {
		printf("original palette size: [n/a], %2d bpp\n",ori_bpp);
	}

    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
        png_set_expand(png_ptr);
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
        png_set_expand(png_ptr);
	/* if (bit_depth == 16)
        png_set_strip_16(png_ptr); */
    if (color_type == PNG_COLOR_TYPE_GRAY ||
        color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
        png_set_gray_to_rgb(png_ptr);

	if (png_get_bKGD(png_ptr, info_ptr, &image_background)) {
		has_bkgd=1;
		bkgd.red=   (unsigned char)image_background->red;
		bkgd.green= (unsigned char)image_background->green;
		bkgd.blue=  (unsigned char)image_background->blue;
		bkgd.alpha=255;
	}

	if (png_get_gAMA(png_ptr, info_ptr, &image_gamma)) {
		has_gama=1;
	}

	if (png_get_sRGB(png_ptr, info_ptr, &srgb_intent)) {
		has_srgb=1;
	}

    png_read_update_info(png_ptr, info_ptr);

	rowbytes=png_get_rowbytes(png_ptr, info_ptr);
	channels=(int)png_get_channels(png_ptr, info_ptr);

	if(channels<3 || channels>4) {
		printf("internal error: channels=%d\n",channels);
		goto abort1;
	}


	image1= (unsigned char*)malloc(height*rowbytes);
	row_pointers = (unsigned char**)malloc(height * sizeof(unsigned char*));
	if(!image1 || !row_pointers) {
		printf("Unable to allocate memory for image\n");
		goto abort1;
	}

	for(i=0;i<(int)height;i++) {
		row_pointers[i] = &image1[rowbytes*i];
	}

	png_read_image(png_ptr, row_pointers);

	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	fclose(fp);

	return 1;

abort1:
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	fclose(fp);
	return 0;
}


int write_new_png(char *fn)
{
    png_structp  png_ptr;
    png_infop  info_ptr;
	png_color palette[256];
	unsigned char trans[256];
	FILE *fp;
	int num_trans;
	int i;
	int new_bit_depth;
	png_color_16 newbackground;


    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL,
      NULL, NULL);
	if (!png_ptr) {
		printf("out of memory\n");
		return 0;
	}

	fp=fopen(fn,"wb");
	if(!fp) {
		printf("can't write to file %s\n",fn);
		png_destroy_write_struct(&png_ptr, NULL);
		return 0;
	}
		

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
		printf("can't create info struct\n");
        goto abort1;
    }

	if (setjmp(png_ptr->jmpbuf)) {
		printf("Error writing file\n");
		goto abort1;
	}

	png_init_io(png_ptr, fp);

	new_bit_depth=8;
	if(pal_used<=16) new_bit_depth=4;
	if(pal_used<=4) new_bit_depth=2;
	if(pal_used<=2) new_bit_depth=1;

	printf("new palette size:        %3d, %2d bpp\n",pal_used,new_bit_depth);

	png_set_IHDR(png_ptr, info_ptr, width, height, new_bit_depth, PNG_COLOR_TYPE_PALETTE,
      interlace_type, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	/* ... set palette colors ... */

	num_trans=0;  /* number of transparent palette entries */

	for(i=0;i<pal_used;i++) {
		palette[i].red=tmppal[i].red;
		palette[i].green=tmppal[i].green;
		palette[i].blue=tmppal[i].blue;

		trans[i]=tmppal[i].alpha;
		if(trans[i]<255) num_trans=i+1;
	}
	png_set_PLTE(png_ptr, info_ptr, palette, pal_used);

	if(num_trans>0) {
		png_set_tRNS(png_ptr, info_ptr, trans, num_trans, 0);
	}

	if(has_gama) png_set_gAMA(png_ptr, info_ptr, image_gamma);
	if(has_srgb) png_set_sRGB(png_ptr, info_ptr, srgb_intent);

	if(has_bkgd) {
		newbackground.index = bkgd_pal;
		png_set_bKGD(png_ptr, info_ptr, &newbackground);
	}
	

	png_write_info(png_ptr, info_ptr);

	png_set_packing(png_ptr);

	/* re-use row_pointers array */
	for(i=0;i<(int)height;i++) {
		row_pointers[i]= &image2[i*width];
	}

	png_write_image(png_ptr, row_pointers);

	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	fclose(fp);

	free(row_pointers);
	free(image2);
	return 1;

abort1:
	png_destroy_write_struct(&png_ptr, NULL);
	fclose(fp);
	return 0;
}


/* set ignorealpha to 1 if you don't care if the alpha value matches
 * (for the background) */
int add_to_palette(tmppal_t *pal, tmppal_t *color, int *pal_used, int ignorealpha)
{
	int used;
	int i;

	used=(*pal_used);
	/* search to see if color is already used */
	for(i=0;i<used;i++) {
		if(pal[i].red==color->red && pal[i].green==color->green &&
			pal[i].blue==color->blue && (pal[i].alpha==color->alpha || ignorealpha))
		{
			return 1;
		}
	}
	if(used>=256) {
		printf("This image can't be converted because it has more than 256 colors.\n");
		return 0;
	}
	pal[used].red=color->red;
	pal[used].green=color->green;
	pal[used].blue=color->blue;
	pal[used].alpha= ignorealpha?255:color->alpha;
	(*pal_used)++;
	return 1;
}

/*
void debug_print_pal(tmppal_t *pal, int used)
{
	int i;
	for(i=0;i<used;i++)
		printf("%d. %d %d %d %d\n",i,pal[i].red,pal[i].green,pal[i].blue,pal[i].alpha);
}
*/

/* Sort the palette to put transparent colors first,
 * then sort approximately dark to light for no particular reason. */
int palsortfunc(const void* p1, const void* p2)
{
	tmppal_t *e1,*e2;
	int s1,s2;

	e1=(tmppal_t*)p1;
	e2=(tmppal_t*)p2;

	if(e1->alpha>e2->alpha) return 1;
	if(e1->alpha<e2->alpha) return -1;
	s1=e1->red+e1->green+e1->blue;
	s2=e2->red+e2->green+e2->blue;
	if(s1>s2) return 1;
	if(s1<s2) return -1;
	return 0;
}


int find_pal_entry(unsigned char r, unsigned char g, unsigned char b, 
				   unsigned char a, int ignorealpha)
{
	int i;

	for(i=0;i<pal_used;i++) {
		if(tmppal[i].red==r && tmppal[i].green==g &&
			tmppal[i].blue==b && (tmppal[i].alpha==a || ignorealpha)) return i;
	}
	return (-1);
}

int make_new_png(void)
{
	tmppal_t thispix;
	unsigned char *p;
	int x,y;
	int palent;

	pal_used=0;

	for(y=0;y<(int)height;y++) {
		for(x=0;x<(int)width;x++) {
			p=&row_pointers[y][x*channels];
				thispix.red=p[0];
				thispix.green=p[1];
				thispix.blue=p[2];
				if(channels==4) thispix.alpha=p[3];
				else thispix.alpha=255;
			if(!add_to_palette(tmppal,&thispix,&pal_used,0)) return 0;
		}
	}

	if(has_bkgd) {
		bkgd.alpha= 255; /* actually not needed */
		add_to_palette(tmppal,&bkgd,&pal_used,1);
	}

	qsort((void*)tmppal,pal_used,sizeof(tmppal_t),palsortfunc);

	/* debug_print_pal(tmppal,pal_used); */

	/* now create the new image */
	image2 = (unsigned char*)malloc(width*height);
	if(!image2) {
		printf("out of memory\n");
		return 0;
	}

	for(y=0;y<(int)height;y++) {
		for(x=0;x<(int)width;x++) {
			p=&row_pointers[y][x*channels];

			palent = find_pal_entry(p[0],p[1],p[2],(unsigned char)((channels==4)?p[3]:255),0);
			if(palent<0) {
				printf("internal error: can't locate palette entry\n");
				return 0;
			}
			image2[y*width + x] = (unsigned char)palent;
		}
	}

	if(has_bkgd) {
		palent = find_pal_entry(bkgd.red,bkgd.green,bkgd.blue,255,1);
		if(palent<0) {
			printf("internal error: can't locale palette entry for bkgd\n");
			return 0;
		}
		bkgd_pal = (unsigned char)palent;
	}

	free(image1);

	return 1;
}

int main(int argc,char **argv)
{

	if(argc!=3) {
		printf("pngrewrite: PNG image palette optimizer\n");
		printf("Usage: %s infile.png outfile.png\n",argv[0]);
		exit(1);
	}

	if(read_png(argv[1])) {
		if(make_new_png()) {
			write_new_png(argv[2]);
		}
	}
	return 0;
}
